/* (C) Copyright 1984,85,86,87 Walter L. Peacock   All Rights Reserved    */
/*  C P Y L T B R O . C                                      */
/* THIS PROGRAM BELONGS TO WALTER L. PEACOCK.  IT IS CONSIDERED A TRADE */
/* SECRET AND IS NOT TO BE DIVULGED OR USED BY PARTIES WHO HAVE NOT       */
/* RECEIVED WRITTEN AUTHORIZATION FROM THE OWNER.                   */
/*---------------------------------------------------------------------------- 
 * cpyltbro() will attempt to copy a cell from the left brother index node 
 * when the current index node underflows (i.e. contains less than MAXCELLS/2 
 * keys) during a delete key.
 *---------------------------------------------------------------------------*/

#include "cbtree.h"

cpyltbro(fdidx, pblk, btparmst)
int fdidx;
struct btidxblk *pblk;
struct btcommo *btparmst;
{
   extern long poptop();
   extern int strlen();
   extern void free();
   extern char *calloc(), *strsave();
   BTIDXBLK *broblk;
   long int lvlnode, cpyloc, pos;
   char *endbroky;
   register int i;

   if(pblk->bwdpage < 2)      /* there is no left brother   */
      return(-1);

   pos = (long) ((pblk->bwdpage - 1L) * btparmst->btidxlen);

   i = 0;     /* lock the record we are about to modify */
   while (lockunlk(fdidx, (long) pos, btparmst->btidxlen, LOCKR) != 1)
      if (++i > HDRTRYS)
         return(-1); 	/* i.e. unable to lock linked list record */

   /* buffer for left brother */
   if( !(broblk = (BTIDXBLK *)calloc(
         sizeof(BTIDXBLK) + btparmst->btcells * SZBTIDX, SZCHAR) ))
      ckerror(- CK_NOMEM, "60");

   getidxr(fdidx, pblk->bwdpage, broblk, btparmst->btidxlen, btparmst->btkeylen);

   if (broblk->cellicnt <= (btparmst->btcells >> 1) ){    /* underflow   */
      lockunlk(fdidx, pos, btparmst->btidxlen, UNLOCKR);
      freekeys(broblk);
      FREE(broblk);
      return(-1);
   }

   /* copy LAST cell from left brother to FIRST position of pblk */

   /* open up space to shift into */
   if (pblk->blkalloc > pblk->cellicnt)
      FREE(pblk->btpage[pblk->cellicnt].skeynme);
   else
      ++pblk->blkalloc; /* we're going to add an allocation from left bro */

   /* shift to make room for cell from left brother */
   blkshft(pblk->btpage + 1, pblk->btpage, pblk->cellicnt++, SZBTIDX);

   /* copy last cell from left brother to our first position */
   blkshft(pblk->btpage, broblk->btpage + --broblk->cellicnt, 1, SZBTIDX);
   --broblk->blkalloc;

   /* compress excess unused blocks over the one just copied */
   blkshft(broblk->btpage + broblk->cellicnt, broblk->btpage +
         broblk->cellicnt + 1, broblk->blkalloc - broblk->cellicnt,
         SZBTIDX);

   putidxr(fdidx, broblk->fwdpage, pblk, btparmst);
   putidxr(fdidx, pblk->bwdpage, broblk, btparmst);

   lockunlk(fdidx, pos, btparmst->btidxlen, UNLOCKR);

   endbroky = strsave(broblk->btpage[broblk->cellicnt-1].skeynme,
         btparmst->btkeylen + 1);

   cpyloc = pblk->bwdpage;

   while ( (lvlnode = poptop() ) != (long) NULL){
      getidxr(fdidx, lvlnode, broblk, btparmst->btidxlen, btparmst->btkeylen);
      i = 0;

      while (broblk->btpage[i++].btptr != cpyloc){
         /* then copied from left tree branch, update upper level */
         if(i >= broblk->cellicnt){
            lvlnode = broblk->bwdpage;
            getidxr(fdidx, lvlnode, broblk, btparmst->btidxlen, btparmst->btkeylen);
            i = 0;
         }
      }

      strcpy(broblk->btpage[--i].skeynme, endbroky);

      putidxr(fdidx, lvlnode, broblk, btparmst);
      cpyloc = lvlnode;
      strcpy(endbroky, broblk->btpage[broblk->cellicnt - 1].skeynme);
   }

   freekeys(broblk);
   FREE(broblk);
   FREE(endbroky);

   return(1);
}

